;****************************************
;*                                      *
;*    MOSTEK Operating System DDT-80    *
;*                                      *
;****************************************
;
; ASCII controls
;
ETX	.EQU	03H		; end of text
LF	.EQU	0AH		; line feed
CR	.EQU	0DH		; carriage return
;
; I/O ports
;
PRPIOD	.EQU	0D0H		; papertape reader PIO data
PRPIOC	.EQU	0D1H		; papertape reader PIO control
PPPIOC	.EQU	0D3H		; papertape punch PIO control
LPPIOC	.EQU	0D7H		; lineprinter PIO control
CTC0	.EQU	0D8H		; CTC channel 0
UDATA	.EQU	0DCH		; UART data port
UCTRL	.EQU	0DDH		; UART control port
SCTRL	.EQU	0DEH		; system control
DCTRL	.EQU	0DFH		; debug control
;
AS	.EQU	0C000H		; Assembler
ED	.EQU	0D48BH		; Editor
ER	.EQU	0D4D9H		; Editor re-entry
EX	.EQU	0DA59H		; Editor ???
TI	.EQU	0DF9BH		;
TO	.EQU	0DF2FH		;
;
; DDT-80 scratchpad area
;
XF000	.EQU	0F000H		;
BRKFLG	.EQU	0FF0CH		; breakpoint flag
BRKADR	.EQU	0FF0DH		; breakpoint address
BRKUSR	.EQU	0FF0FH		; user code at breakpoint address
INTFLG	.EQU	0FF12H		; enable or disable interrupt
REGFLG	.EQU	0FF13H		; display only PC&AF or all registers
OPR1	.EQU	0FF14H		; operand 1
OPR2	.EQU	0FF16H		; operand 2
OPR3	.EQU	0FF18H		; operand 3
OPFLG	.EQU	0FF1AH		; nbr of operands
YFF1B	.EQU	0FF1BH		;
CMDLTR	.EQU	0FF1CH		; command letter
YFF1D	.EQU	0FF1DH		;
SYSEXT	.EQU	0FF1FH		; system extension
PRFLG	.EQU	0FF21H		; papertape reader flag
PPFLG	.EQU	0FF22H		; papertape punch flag
LPFLG	.EQU	0FF23H		; lineprinter flag
;
CIADR	.EQU	0FF27H		; console input
COADR	.EQU	0FF29H		; console output
OIADR	.EQU	0FF2BH		; object input
OOADR	.EQU	0FF2DH		; object output
SIADR	.EQU	0FF2FH		; source input
SOADR	.EQU	0FF31H		; source output
;
USRMNE	.EQU	0FF33H		; user mnemonics table
USRSTK	.EQU	0FF90H		; user stack
;
TMPSP	.EQU	0FFB4H		; temp loc for SP
;
SPLOC	.EQU	0FFE6H		;
IYLOC	.EQU	0FFE8H		;
IXLOC	.EQU	0FFEAH		;
LALOC	.EQU	0FFECH		;
HALOC	.EQU	0FFEDH		;
EALOC	.EQU	0FFEEH		;
DALOC	.EQU	0FFEFH		;
CALOC	.EQU	0FFF0H		;
BALOC	.EQU	0FFF1H		;
FALOC	.EQU	0FFF2H		;
AALOC	.EQU	0FFF3H		;
LLOC	.EQU	0FFF4H		;
HLOC	.EQU	0FFF5H		;
ELOC	.EQU	0FFF6H		;
DLOC	.EQU	0FFF7H		;
CLOC	.EQU	0FFF8H		;
BLOC	.EQU	0FFF9H		;
IFLOC	.EQU	0FFFAH		;
ILOC	.EQU	0FFFBH		;
ALOC	.EQU	0FFFDH		;
FLOC	.EQU	0FFFCH		;
PCLOC	.EQU	0FFFEH		;
;
	.ORG	0E000H			
;
	JP	START		;
;
START:	IN	A,(DCTRL)	;
	LD	HL,USRSTK	;
	LD	(SPLOC),HL	;
	XOR	A		;
	LD	(BRKFLG),A	; no breakpoint
	LD	A,05H		;
	OUT	(CTC0),A	; initialize CTC
	LD	A,03H		;
	OUT	(SCTRL),A	;
;
; compute baud rate
;
	LD	C,SCTRL		;
	LD	DE,1		;
WAITH:	IN	A,(C)		; wait for high
	JP	P,WAITH		;
WAITL:	INC	DE		; count loops
	IN	A,(C)		;
	JP	M,WAITL		; until low
	LD	SP,CTCTAB-1	; adress CTC table
AE02A:	INC	SP		; 
	POP	HL		; get threshold
	SCF			;
	SBC	HL,DE		; compare pulse width
	JR	C,AE02A		;
	DEC	SP		;
	POP	AF		; get constant
	OUT	(CTC0),A	; set CTC time constant
	LD	HL,TT		; teletype typehead 
	CP	020H		; or
	JR	NZ,SETCO	;
	LD	HL,ST		; Silent printer
SETCO:	LD	(COADR),HL	; set console output
	CP	057H		;
	LD	A,0D8H		; 2 stop bits, no parity, 8 data bits
	JR	Z,STUART		;
	LD	A,058H		; 1 stop bit, no parity, 8 data bits
STUART:	OUT	(UCTRL),A	; set UART control
	JP	INIT		; go ahead
;
; table for CTC initialization
;
CTCTAB:	.WORD	000DH		; threshold
	.BYTE	001H		; time constant
;
	.WORD	001AH		;
	.BYTE	002H		;
;
	.WORD	0034H		;
	.BYTE	004H		;
;
	.WORD	0068H		;
	.BYTE	008H		;
;
	.WORD	00D0H		;
	.BYTE	010H		;
;
	.WORD	01A0H		;
	.BYTE	020H		;
;
	.WORD	7FFFH		;
	.BYTE	057H		;
;
	.WORD	13FFH		;
;
; NMI
;
	JP	SAVERG		;
;
; save registers  into scratchpad
;
SAVERG:	LD	(SPLOC),SP	; save stack pointer
	LD	SP,00000H	; adjust stack pointer
	PUSH	AF		;
	PUSH	AF		;
	OR	A		;
SAVRG1:	IN	A,(DCTRL)	; reset debug flag
	LD	A,I		;
	PUSH	AF		;
	DI			;
	PUSH	BC		;
	PUSH	DE		;
	PUSH	HL		;
	EXX			;
	EX	AF,AF'		;
	PUSH	AF		;
	PUSH	BC		;
	PUSH	DE		;
	PUSH	HL		;
	LD	A,(IFLOC)	;
	AND	004H		;
	LD	(IFLOC),A	;
	EXX			;
	EX	AF,AF'		;
	PUSH	IX		;
	PUSH	IY		;
	JR	NC,AE0A8	; breakpoint routine?
	LD	HL,BRKUSR	; yes, restore user code
	LD	DE,(BRKADR)	; to breakpoint address
	LD	BC,3		;
	LDIR			;
	LD	HL,(BRKADR)	;
	LD	A,081H		;
	LD	(BRKFLG),A	;
	JR	AE0B1		;
;
AE0A8:	LD	SP,(SPLOC)	;
	POP	HL		;
	LD	(SPLOC),SP	;
AE0B1:	LD	(PCLOC),HL	;
	LD	SP,SPLOC	;
	JP	AE621		;
;
; breakpoint routine
;
BRKRTN:	LD	(SPLOC),SP	; save stack pointer
	LD	SP,00000H	; adjust stack pointer
	PUSH	AF		;
	PUSH	AF		;
	SCF			; signalize breakpoint routine
	JR	SAVRG1		; and save registers
;
RSTRG0:	LD	A,00H		; signalize execute
	JR	RSTRGS		;
;
RSTRG1:	LD	A,01H		; signalize breakpoint
;
; restore registers
;
RSTRGS:	LD	(INTFLG),A	;
	LD	SP,(SPLOC)	;
	LD	HL,(PCLOC)	;
	PUSH	HL		;
	LD	HL,(FLOC)	;
	PUSH	HL		;
	LD	(TMPSP),SP	; save stack pointer
	LD	SP,IYLOC	;
	POP	IY		;
	POP	IX		;
	EXX			;
	EX	AF,AF'		;
	POP	HL		;
	POP	DE		;
	POP	BC		;
	POP	AF		;
	EXX			;
	EX	AF,AF'		;
	POP	HL		;
	POP	DE		;
	POP	BC		;
	POP	AF		;
	LD	I,A		;
	LD	SP,(TMPSP)	; restore stack pointer
	LD	A,(INTFLG)	;
	JP	PE,AE103	; ???
	OUT	(DCTRL),A	;
	POP	AF		;
	DI			; disable interrupt
	RET			;
;
AE103:	OUT	(DCTRL),A	;
	POP	AF		;
	EI			; enable interrupt
	RET			;
;
INIT:	LD	HL,TK		; teletype keyboard 
	LD	(CIADR),HL	;  is console input
	LD	HL,EXTRET	; shorten
	LD	(SYSEXT),HL	;  system extension
	XOR	A		; clear
	LD	(IFLOC),A	;  user interrupt flag
	LD	A,080H		; clear
	LD	(USRMNE),A	;  user mnemonics
	LD	SP,SPLOC	;
	CALL	DELAY		;
	IN	A,(UDATA)	; read UART
	IM	2		; set interrupt mode
;
; main loop with new line
;
MLOOP0:	LD	E,00H		; select console
	CALL	CRLF		; go to new line
;
; main loop w/o new line
;
MLOOP1:	LD	DE,'.'*256	; write prompt on console
	CALL	WRCHR		;
	CALL	ECHO		; get command
	CP	'.'		;
	JR	Z,MLOOP0	;
	LD	HL,CMDLTR	;
	LD	(HL),D		; save command letter
	CALL	SPACE		; write a blank
	CALL	DSCAN		; get operands
	LD	HL,(SYSEXT)	;
	JP	(HL)		; start system extension
;
; return from unsuccessful system extension
;
EXTRET:	IN	A,(SCTRL)	;
	BIT	3,A		;
	JR	Z,LODOPR	;
	JP	XF000		;
;
LODOPR:	LD	A,(CMDLTR)	; load operands
	LD	HL,(OPR2)	;
	LD	DE,(OPR1)	;
;
; M command: display and update memory
;
	CP	'M'		;
	JP	NZ,PCMND	;
	EX	DE,HL		;
	LD	E,001H		;
	LD	A,(OPFLG)	;
	CP	001H		;
	JR	Z,AE195		;
AE169:	PUSH	HL		;
	POP	BC		;
	PUSH	HL		;
	LD	HL,(OPR2)	;
	OR	A		;
	SBC	HL,BC		;
	LD	BC,0000FH	;
	OR	A		;
	SBC	HL,BC		;
	LD	B,010H		;
	JR	Z,AE182		;
	JP	NC,AE184	;
	LD	A,L		;
	ADD	A,B		;
	LD	B,A		;
AE182:	SET	5,E		;
AE184:	POP	HL		;
	CALL	PADDO		; print HL
	CALL	PRTMEM		; print memory
	CALL	CRLF		;
	BIT	5,E		;
	JP	NZ,MLOOP1	;
	JR	AE169		;
;
AE195:	CALL	AE5F7		;
	BIT	6,E		;
	JR	Z,AE1AE		;
	PUSH	HL		;
	EXX			;
	POP	IX		;
	LD	L,(IX+000H)	;
	LD	H,(IX+001H)	;
	LD	E,000H		;
	CALL	AE5F7		;
	EXX			;
	JR	AE1B2		;
;
AE1AE:	LD	A,(HL)		;
	CALL	PACCBL		;
AE1B2:	PUSH	DE		;
	LD	(YFF1D),HL	;
	CALL	DSCAN		;
	LD	HL,(YFF1D)	;
	POP	DE		;
	LD	A,(YFF1B)	;
	CP	02EH		;
	JP	Z,MLOOP1		;
	LD	A,(OPFLG)	;
	AND	003H		;
	JR	Z,AE1EC		;
	LD	A,(OPR1)	;
	LD	(HL),A		;
	INC	HL		;
	BIT	6,E		;
	JR	Z,AE1DA		;
	LD	A,(OPR1+1)	;
	LD	(HL),A		;
	INC	HL		;
AE1DA:	LD	A,(YFF1B)	;
	CP	00DH		;
	JR	Z,AE195		;
AE1E1:	DEC	HL		;
	DEC	HL		;
	CALL	USRTAB		; lookup menonics
	JP	M,AE1EA		;
AE1E9:	INC	HL		;
AE1EA:	JR	AE195		;
;
AE1EC:	LD	A,(YFF1B)	;
	CP	'^'		;
	JR	Z,AE1E1		;
	INC	HL		;
	BIT	6,E		;
	JR	NZ,AE1E9	;
	JR	AE195		;
;
; P command: display and/or modify port
;
PCMND:	CP	'P'		;
	JR	NZ,SCMND	;
	LD	A,(OPR1)	; get port address
	LD	C,A		;
PLOOP:	LD	E,001H		;
	LD	A,C		;
	CALL	PACCBL		; write port address
	IN	A,(C)		;
	CALL	PACCBL		; write input byte
	PUSH	BC		;
	CALL	DSCAN		; 
	POP	BC		;
	LD	A,(YFF1B)	;
	CP	'.'		;
	JP	Z,MLOOP1		;
	LD	H,A		;
	LD	A,(OPFLG)	; any operands?
	AND	A		;
	JR	Z,AE22E		; no ,skip
	LD	A,(OPR1)	;
	OUT	(C),A		; output operand
	LD	A,'^'		;
	CP	H		;
	JR	Z,PLOOP		;
PNEXT:	INC	C		; next port address
	JR	PLOOP		; repeat
;
AE22E:	LD	A,'^'		;
	CP	H		;
	JR	NZ,PNEXT	;
	DEC	C		;
	JR	PLOOP		;
;
; S command: single step
;
SCMND:	CP	'S'		;
	JR	NZ,BCMND	;
	LD	A,(OPFLG)	; any operands?
	AND	A		;
	JR	Z,AE244		; no, skip
	EX	DE,HL		;
	LD	(PCLOC),HL	;
AE244:	LD	A,(OPR3)	;
	LD	(REGFLG),A	;
	LD	A,(OPR2)	;
	AND	A		;
	JR	NZ,AE251	;
AE250:	INC	A		;
AE251:	LD	(OPR2),A	;
	JP	RSTRG1		; restore registers
;
; B command: set/reset breakpoint
;
BCMND:	CP	'B'		;
	JR	NZ,ECMND	;
	PUSH	DE		;
	LD	A,(BRKFLG)	;
	AND	A		;
	JR	Z,SETBRK		;
	LD	DE,(BRKADR)	;
	LD	HL,BRKUSR	;
	LD	BC,3		;
	LDIR			; restore user code
	XOR	A		;
	LD	(BRKFLG),A	;
SETBRK:	LD	A,(OPFLG)	; any operands? 
	AND	A		;
	POP	HL		;
	JR	Z,MLOOP4	; no, exit
	LD	(BRKADR),HL	; store breakpoint address
	PUSH	HL		;
	LD	DE,BRKUSR	;
	LD	BC,3		;
	PUSH	BC		;
	LDIR			; save user code
	LD	HL,BRKCOD	; 
	POP	BC		;
	POP	DE		;
	LDIR			; load breakpoint code
	LD	A,(OPR2)	;
	LD	(REGFLG),A	;
	LD	A,01H		; signalize
	LD	(BRKFLG),A	; breakpoint
MLOOP4:	JP	MLOOP1		;
;
; breakpoint code
;
BRKCOD:	JP	BRKRTN		;
;
; E command: start a program
;
ECMND:	CP	'E'		;
	JR	NZ,DCMND	;
	LD	A,(OPFLG)	; any operands
	AND	A		;
	JR	Z,AE2AC		;
	LD	(PCLOC),DE	; set new PC value
AE2AC:	JP	RSTRG0		; restore registers
;
; D command: dump memory
;
DCMND:	CP	'D'		;
	JR	NZ,LCMND	;
	XOR	A		;
	SBC	HL,DE		; compute length
	INC	HL		;
	PUSH	HL		; and save it
	LD	E,00BH		; set I/O channel
	CALL	TRAILR		; write trailer
DLOOP:	LD	D,':'		; start of record
	CALL	WRCHR		;
	POP	HL		; get length back
	LD	BC,16		; subtract record length
	XOR	A		;
	SBC	HL,BC		;
	JR	NC,AE2D4	; short record?
	ADD	HL,BC		; yes, 
	ADD	A,L		;
	LD	B,A		;
	LD	L,00H		;
	JR	Z,DEND		;
	JR	AE2D5		;
;
AE2D4:	LD	B,C		;
AE2D5:	PUSH	HL		;
	LD	C,00H		;
	LD	A,B		;
	CALL	PTWDG		;
	LD	HL,(OPR1)	; get memory pointer
	LD	A,H		;
	CALL	PTWDG		;
	LD	A,L		;
	CALL	PTWDG		;
	XOR	A		;
	CALL	PTWDG		;
MNEXT:	LD	A,(HL)		;
	CALL	PTWDG		;
	INC	HL		;
	DJNZ	MNEXT		;
	CALL	WRCHK		;
	LD	(OPR1),HL	; set memory pointer
	JR	DLOOP		;
;
; write the end record
;
DEND:	LD	B,3		; write 3 null bytes
DEND1:	XOR	A		; 
	LD	C,A		;
	CALL	PTWDG		;
	DJNZ	DEND1		;
	LD	A,001H		; write last record type
	CALL	PTWDG		;
	CALL	WRCHK		; write checksum
	CALL	TRAILR		; write trailer
MLOOP2:	JP	MLOOP0		; and go back
;
; convert to ASCII, output and checksum
;
PTWDG:	PUSH	AF		;
	ADD	A,C		; add to checksum
	LD	C,A		;
	POP	AF		;
	JP	PACC		; convert & output
;
; L command: load memory
;
LCMND:	CP	'L'		;
	JR	NZ,CCMND	;
	LD	E,0AH		;
	LD	A,(OPFLG)	; any operands? 
	AND	A		;
	JP	NZ,EX		; no, jump to editor
LLOOP:	CALL	RDCHR		; wait for 
	SUB	':'		; start of record
	JR	NZ,LLOOP	;
	LD	C,A		;
	CALL	RDBYTE		; get record length
	LD	B,A		;
	CALL	RDBYTE		; get high address
	LD	H,A		;
	CALL	RDBYTE		; get low address
	LD	L,A		;
	CALL	RDBYTE		; get record type
	DEC	A		;
	PUSH	AF		;
	JR	Z,CHCKSM		; end record?
LNEXT:	CALL	RDBYTE		; read next byte
	LD	(HL),A		; and fill memory
	INC	HL		; increment pointer
	DJNZ	LNEXT		; 
CHCKSM:	CALL	RDBYTE		; get checksum
	XOR	A		;
	ADD	A,C		;
	JR	Z,AE35A		; checksum o.k.?
	LD	E,00H		; no,
	LD	A,H		; write address
	CALL	PACC		;
	LD	A,L		;
	CALL	PACC		;
	LD	E,02H		;
AE35A:	POP	AF		; end of file?
	JR	NZ,LLOOP	; no, loop
	CALL	RDCHR		;
	JR	MLOOP2		; back to command loop
;
; read a byte and add to checksum
;
RDBYTE:	CALL	RDCHR		; next char
	CALL	ASBIN		; to bin
	RLCA			; high nibble
	RLCA			;
	RLCA			;
	RLCA			;
	PUSH	BC		; save checksum
	LD	C,A		; save high nibble
	CALL	RDCHR		; next char
	CALL	ASBIN		; to bin
	OR	C		; complete byte
	POP	BC		; restore checksum
	PUSH	AF		; save byte
	ADD	A,C		; add to checksum
	LD	C,A		;
	POP	AF		;
	RET			;
;
; send trailer
;
TRAILR:	PUSH	BC		;
	LD	B,50H		; 80
	LD	D,00H		; nulls
TRLR1:	CALL	WRCHR		; to write
	DJNZ	TRLR1		;
	POP	BC		;
	RET			;
;
; C command: copy memory
;
CCMND:	CP	'C'		;
	JR	NZ,RCMND	;
	OR	A		;
	SBC	HL,DE		; compute block length
	INC	HL		;
	LD	B,H		; set counter
	LD	C,L		;
	EX	DE,HL		; save end address
	LD	DE,(OPR3)	; determine direction
	OR	A		;
	SBC	HL,DE		;
	JR	NC,UPLOAD	;
	LD	HL,(OPR3)	; get destination address
	ADD	HL,BC		; add block length
	DEC	HL		;
	LD	D,H		; set new destination address
	LD	E,L		;
	LD	HL,(OPR2)	; new start address
	LDDR			; download
	JR	MLOOP3		; back to main loop
;
UPLOAD:	LD	HL,(OPR1)	; load start address
	LDIR			; upload
MLOOP3:	JP	MLOOP1		; back to main loop
;
; R command: display user registers
;
RCMND:	CP	'R'		;
	JR	NZ,MLOOP3	; no further commands
	LD	A,(OPR1)	;
	AND	A		;
	JR	Z,NOHDR		;
	LD	HL,REGHDR	; print header
	CALL	PTXT		;
NOHDR:	CALL	PRTALL		;
	JP	MLOOP0		; back to main loop
;
; print a string until ETX
;
PTXT:	LD	D,(HL)		; get next character
	LD	A,D		;
	CP	ETX		; end of text ?
	RET	Z		; yes, return
	CALL	WRCHR		; write character
	INC	HL		; increment pointer
	JR	PTXT		; & loop
;
; register display header
;
REGHDR:	.BYTE	" PC   AF  I IF";
	.BYTE	"  BC   DE   HL";
	.BYTE	"  A",27H	;
	.BYTE	"F",27H		;
	.BYTE	" B",27H	;
	.BYTE	"C",27H		;
	.BYTE	" D",27H 	;
	.BYTE	"E",27H  	;
	.BYTE	" H",27H 	;
	.BYTE	"L",27H		;
	.BYTE	"  IX   IY   SP";  	 
	.BYTE	CR,LF,ETX	;
;
; DDT-80 scan routine
;
DSCAN:	XOR	A		; clear operands
	LD	HL,OPR1		;
	PUSH	HL		;
	POP	IY		; IY points to first operand
	LD	(HL),A		; clear all operands
	LD	BC,5		;
	LD	DE,OPR1+1	;
	LDIR			;
	LD	E,A		; set I/O channel to console
	LD	(OPFLG),A	; clear number of operands
	EXX			;
	LD	D,A		;
AE42A:	LD	B,0		;
AE42C:	XOR	A		;
	LD	H,A		;
	LD	L,A		;
AE42F:	LD	E,D		;
	EXX			;
	CALL	ECHO		; get a character
	EXX			;
	LD	D,A		;
	CALL	AORN		; num or alpha?
	LD	A,D		;
	JR	NZ,AE44A	;
	CALL	ASBIN		;
	ADD	HL,HL		;
	ADD	HL,HL		;
	ADD	HL,HL		;
	ADD	HL,HL		;
	ADD	A,L		;
	LD	L,A		;
	JR	NC,AE42F	;
	INC	H		;
AE448:	JR	AE42F		;
;
AE44A:	CP	'$'		;
	JR	NZ,AE454	;
	LD	HL,(YFF1D)	;
	INC	HL		;
	JR	AE42F		;
;
AE454:	PUSH	DE		;
	PUSH	HL		;
	PUSH	IY		;
	POP	DE		;
	EX	DE,HL		;
	XOR	A		;
	ADD	A,B		;
	JR	Z,AE467		;
	LD	A,(HL)		;
	SUB	E		;
	LD	(HL),A		;
	INC	HL		;
	LD	A,(HL)		;
AE463:	SBC	A,D		;
	LD	(HL),A		;
	JR	AE46E		;
;
AE467:	LD	A,(HL)		;
	ADD	A,E		;
	LD	(HL),A		;
	INC	HL		;
	LD	A,(HL)		;
	ADC	A,D		;
	LD	(HL),A		;
AE46E:	POP	HL		;
	POP	DE		;
	LD	A,D		;
	CP	'+'		;
	JR	NZ,AE479	;
	LD	B,0		;
	JR	AE42C		;
;
AE479:	CP	'-'		;
	JR	NZ,AE481	;
	LD	B,001H		;
AE47F:	JR	AE42C		;
;
AE481:	CP	' '		;
	JR	Z,AE491		;
	CP	','		;
	JR	Z,AE491		;
	CP	CR		;
	JR	Z,AE491		;
	CP	'^'		;
	JR	NZ,AE4CC	;
AE491:	LD	A,(OPFLG)	;
	CP	ETX		;
	JR	Z,AE4A4		;
	INC	A		;
	LD	(OPFLG),A	;
	CP	ETX		;
	JR	Z,AE4A4		;
	INC	IY		;
	INC	IY		;
AE4A4:	LD	A,D		;
	EXX			;
	CP	CR		;
	JR	Z,AE4B7		;
	CP	'^'		;
	JR	Z,AE4B2		;
	EXX			;
	JP	AE42A		;
;
AE4B2:	LD	D,CR		;
	CALL	WRCHR		;
AE4B7:	CALL	PRLF		;
	EXX			;
	LD	A,E		;
	AND	A		;
	JR	NZ,AE4C6	;
	LD	A,(OPFLG)	;
	DEC	A		;
	LD	(OPFLG),A	;
AE4C6:	LD	A,D		;
	LD	(YFF1B),A	;
	EXX			;
	RET			;
;
AE4CC:	CP	':'		;
	JR	Z,AE4EF		;
	CP	'.'		;
	JR	NZ,AE4DB	;
AE4D4:	XOR	A		;
	LD	(CMDLTR),A	;
	EXX			;
	JR	AE4B2		;
;
AE4DB:	CP	'='		;
	JR	NZ,AE4D4	;
	LD	A,(IY+001H)	;
	EXX			;
	CALL	PACC		;
	LD	A,(IY+000H)	;
	CALL	PACC		;
	EXX			;
	JR	AE47F		;
;
AE4EF:	EXX			;
	CALL	ECHO		;
	EXX			;
	LD	L,A		;
	EXX			;
	CALL	ECHO		;
	EXX			;
	LD	D,A		;
	LD	E,A		;
	CP	'.'		;
	JR	Z,AE4D4		;
	CALL	AORN		;
	JR	Z,AE507		;
	LD	D,' '		;
AE507:	LD	H,D		;
	CALL	LKPUSR		; lookup user mnemonics
	JR	Z,MNEFND	;
	CALL	LKPSTD		; lookup standard mnemonics
	JR	Z,MNEFND	;
	XOR	A		; mnemonic not found
	LD	H,A		; clear address
	LD	L,A		;
AE515:	JP	AE42F		; back to scan routine
;
MNEFND:	LD	A,E		;
	CALL	AORN		;
	LD	D,E		;
	JR	Z,AE515		;
	JP	AE454		;
;
RDCHR:	LD	A,E		; get channel #
	AND	0FEH		; reset bit 0 (input)
	JR	CHRIO		;
;
WRCHR:	LD	A,E		; get channel #
	OR	001H		; set bit 0 (output)
;
; character I/O
; get driver address from tab and call it
;
CHRIO:	PUSH	HL		; save HL
	LD	HL,CHRRET	; get return address
	PUSH	HL		;  and push it
	AND	007H		; mask channel #
	SLA	A		; double value
	ADD	A,CIADR%256	; add to driver tab address
	LD	L,A		;
	XOR	A		;
	ADC	A,CIADR/256	;
	LD	H,A		;
	LD	A,(HL)		; get address
	INC	HL		;
	LD	H,(HL)		;
	LD	L,A		;
	JP	(HL)		; and jump
;
CHRRET:	POP	HL		; restore HL
	RET			; & return
;
; translate mnemonic to address
;
LKPSTD:	PUSH	DE		;
	LD	DE,MNETAB	; lookup standard mnemonics
	JR	SRCMNE		;
;
LKPUSR:	PUSH	DE		;
	LD	DE,USRMNE	; lookup user mnemonics
SRCMNE:	EX	DE,HL		;
NXTMNE:	LD	A,(HL)		; get next mnemonic
	CP	080H		; end of table?
	JR	Z,MNERET	; yes, exit 
 	CP	E		; compare first charcater
	JR	NZ,INCHL4	; no match, skip to next menmonic
	INC	HL		;
	LD	A,(HL)		;
	AND	07FH		;
	CP	D		; compare second character
	JR	NZ,INCHL3	; no match, skip to next menmonic
	INC	HL		; mnemonic found
	LD	E,(HL)		; get address
	INC	HL		;
	LD	D,(HL)		;
	XOR	A		; signalize success
MNERET:	POP	HL		; return
	EX	DE,HL		;
	OR	A		; set zero flag   
	RET			;
;
; skip to next menmonic
;
INCHL4:	INC	HL		; adjust pointer
INCHL3:	INC	HL		;
	INC	HL		;
	INC	HL		;
	JR	NXTMNE		; try next mnemonic
;
; alphanumeric determination
;
AORN:	CP	027H		;
	JR	Z,AE57E		;
	CP	030H		;
	JR	C,AE580		;
	CP	03AH		;
	JR	C,AE57E		;
	CP	040H		;
	JR	C,AE580		;
	CP	05BH		;
	JR	NC,AE580	;
AE57E:	XOR	A		;
	RET			;
;
AE580:	XOR	A		;
	INC	A		;
	RET			;
;
; convert ASCII hex character to binary
;
ASBIN:	SUB	'0'		; subtract ASCII 0
	CP	10		; <10
	RET	M		; yes, return
	SUB	7		; otherwise subtract 7
	RET			; & return
;
; output accumulator as two hex characters
;
PACC:	PUSH	AF		;
	RRCA			;
	RRCA			;
	RRCA			;
	RRCA			;
	CALL	PRVAL		; 
	POP	AF		;
	JP	PRVAL		;
;
; read a character and write it back
;
ECHO:	CALL	RDCHR		;
	JR	WRCHR1		;
;
; send a carriage return & line feed
;
CRLF:	LD	D,CR		;
	CALL	WRCHR		;
PRLF:	LD	D,LF		;
WRCHR1:	JR	WRCHR2		;
;
SPACE:	LD	D,' '		;
WRCHR2:	JP	WRCHR		;
;
; print accu and a blank
;
PACCBL:	CALL	PACC		;
	JR	SPACE		;
;
; convert to ASCII and output
;
PRVAL:	AND	0FH		;
	ADD	A,90H		;
	DAA			;
	ADC	A,40H		;
	DAA			;
	LD	D,A		;
	JR	WRCHR2		;
;
; translate address to mnemonic
;
USRTAB:	LD	IY,USRMNE	;
	LD	B,2		; two tables
	JR	LOOKUP		;
;
STDTAB:	LD	IY,MNETAB	;
LOOKUP:	LD	C,00H		;
	LD	A,(IY+0)	;
	CP	080H		; end of table?
	JR	Z,NXTTAB	; yes, exit
	INC	IY		;
	INC	IY		;
	LD	A,(IY+0)	;
	CP	L		;
	JR	NZ,AE5ED	;
	INC	IY		;
	LD	A,(IY+0)	;
	CP	H		;
	JR	NZ,AE5EF	;
	LD	A,(IY-3)	;
	LD	B,A		;
	LD	A,(IY-2)	;
	LD	C,A		;
	AND	080H		;
	INC	A		;
	RET			;
;
AE5ED:	INC	IY		;
AE5EF:	INC	IY		;
	JR	LOOKUP		;
;
NXTTAB:	DJNZ	STDTAB		; both tables done? 
	XOR	A		; yes,
	RET			; return
;
AE5F7:	RES	6,E		;
	CALL	USRTAB		; lookup menmonics
	JP	P,AE601		;
	SET	6,E		;
AE601:	JP	NZ,AE60F	;
;
; convert HL to ASCII and output
;
PADDO:	LD	A,H		;
	CALL	PACC		;
	LD	A,L		;
	CALL	PACC		;
SPACE1:	JP	SPACE		;
;
AE60F:	LD	D,':'		;
	CALL	WRCHR		;
	LD	D,B		;
	CALL	WRCHR		;
	LD	D,C		;
	CALL	WRCHR		;
	CALL	SPACE		;
	JR	SPACE1		;
;
AE621:	LD	E,001H		;
	CALL	PRPCAF		;
	LD	A,(BRKFLG)	;
	AND	A		;
	LD	A,00H		;
	LD	(BRKFLG),A	;
	JP	M,MLOOP0	;
	LD	A,(OPR2)	;
	DEC	A		;
	JR	Z,AE63B		;
	DEC	A		;
	JR	AE64C		;
;
AE63B:	LD	E,00H		;
	CALL	RDCHR		;
	LD	A,'.'		;
	CP	D		;
	JP	Z,MLOOP0	;
	LD	A,CR		;
	CP	D		;
	JR	NZ,AE63B	;
	XOR	A		;
AE64C:	PUSH	AF		;
	CALL	CRLF		;
	POP	AF		;
	JP	AE250		;
;
PRPCAF:	LD	A,(REGFLG)	;
	RRA			;
	LD	B,2		; print only PC & AF
	JR	NC,PRREGS	;
PRTALL:	LD	B,13		; print all registers
PRREGS:	LD	HL,PCLOC+1	;
RGLOOP:	LD	A,(HL)		;
	CALL	PACC		; print high byte
	DEC	HL		;
	LD	A,(HL)		;
	CALL	PACCBL		; print low byte
	DEC	HL		;
	DJNZ	RGLOOP		;
	RET			;
;
; print <B> memory bytes
;
PRTMEM:	LD	A,(HL)		; get next byte
	CALL	PACCBL		; print it
	INC	HL		; increment pointer
	DJNZ	PRTMEM		; repeat until done
	RET			;
;
; write checksum
;
WRCHK:	SUB	A		; clear accumulator
	SUB	C		; negate checksum
	CALL	PTWDG		;  and write it
	JP	CRLF		; write CR/LF and return
;
; Silent 700 driver
;
ST:	SET	4,E		;
;
; terminal typehead driver
;
TT:	IN	A,(UCTRL)	; get UART status
	BIT	7,A		; transmitter buffer empty?
	JR	NZ,TTSEND	; yes, go send data
	BIT	7,E		; immediate return?
	RET	NZ		; yes, return
	JR	TT		; otherwise wait
;
TTSEND:	LD	A,D		;
	OUT	(UDATA),A	; send data
	BIT	4,E		;
	JR	Z,TTRET		;
	CP	CR		;
	JR	NZ,TTRET	; return w/o delay
DELAY:	PUSH	BC		;
	LD	BC,07487H	;
DELAY1:	DEC	C		;
	JR	NZ,DELAY1	;
	DJNZ	DELAY1		;
	POP	BC		;
TTRET:	RES	7,E		;
	RES	4,E		;
	RET			;
;
; teletype reader driver
;
TR:	LD	A,07H		;
	OUT	(SCTRL),A	;
TRWAIT:	IN	A,(SCTRL)	;
	BIT	7,A		;
	JR	Z,TRWAIT	;
	LD	A,03H		;
	OUT	(SCTRL),A	;
;
; terminal keybord driver
;
TK:	IN	A,(UCTRL)	;
	BIT	6,A		; char available?
	JR	NZ,TKGET	; yes, get it
	BIT	7,E		; immediate return?
	RET	NZ		; yes
	JR	TK		; otherwise wait
;
TKGET:	IN	A,(UDATA)	; get character
	RES	7,E		; reset immediate bit
	RES	7,A		; reset parity
	LD	D,A		; 
	RET			;
;
; paper tape reader driver
;
PR:	PUSH	HL		;
	PUSH	BC		;
	LD	BC,PROFF+PRPIOC	;
	LD	A,04FH		;
	LD	HL,PRFLG	;
	BIT	3,E		; init?
	JR	Z,PRENI		; no, skip
	CALL	INITP		; init PIO
	IN	A,(PRPIOD)	; 
PRENI:	EI			;
PRWAIT:	BIT	7,(HL)		;
	JR	NZ,PROUT	;
	BIT	7,E		;
	JR	NZ,IORET2	;
	JR	PRWAIT		;
;
PROUT:	LD	(HL),000H	; clear flag
	IN	A,(PRPIOD)	; get character
	CPL			;
	RES	7,E		; reset imm flag
	RES	7,A		; reset parity
	LD	D,A		;
IORET2:	JR	IORET		;
;
; line printer driver
;
LP:	PUSH	HL		;
	LD	HL,LPFLG	;
	PUSH	BC		;
	LD	BC,LPOFF+LPPIOC	;
	JR	PP1		;
;
; paper tape punch driver
;
PP:	PUSH	HL		;
	LD	HL,PPFLG	;
	PUSH	BC		;
	LD	BC,PPOFF+PPPIOC	;
PP1:	PUSH	AF		;
	BIT	3,E		; init necessary
	JR	Z,PPENI		; no
	LD	A,00FH		; otherwise
	CALL	INITP		; init PIO
	JR	PPOUT		;
;
PPENI:	EI			; enable interrupt
PPWAIT:	BIT	7,(HL)		; check flag
	JR	NZ,PPOUT	;
	BIT	7,E		; immediate return?
	JR	NZ,IORET1	; yes
	JR	PPWAIT		; otherwise wait
;
PPOUT:	LD	(HL),00H	; reset flag
	LD	A,D		; load character
	DEC	BC		; address data port
	OUT	(C),A		; send character
	RES	7,E		; reset immediate flag
IORET1:	POP	AF		; restore registers
IORET:	POP	BC		;
	POP	HL		;
	RES	3,E		; reset init flag
	RET			; & return
;
	.BYTE	0FFH		;
;
; interrupt table
;
INTTAB:	
PROFF	.EQU	($-INTTAB)*256	;
	.WORD	INTPR		;
PPOFF	.EQU	($-INTTAB)*256	;
	.WORD	INTPP		;
LPOFF	.EQU	($-INTTAB)*256	;
	.WORD	INTLP		;
;
; initialize PIO
;
INITP:	DI			;
	OUT	(C),A		;
	LD	A,INTTAB%256	;
	ADD	A,B		; add offset
	OUT	(C),A		;
	LD	A,083H		; control byte
	OUT	(C),A		;
	LD	(HL),00H	; reset flag
	LD	A,INTTAB/256	; 
	LD	I,A		;
	EI			; enable interrupt
	RET			;
;
; interrupt sevice routines
;
INTPR:	PUSH	AF		; save accumulator
	LD	A,0FFH		; set flag
	LD	(PRFLG),A	;
;
INTRET:	POP	AF		; restore accumulator
	EI			; enable interrupt
	RETI			; & return
;
INTPP:	PUSH	AF		; save accumulator
	LD	A,0FFH		; set flag
	LD	(PPFLG),A	;
	JR	INTRET		; & return
;
INTLP:	PUSH	AF		; save accumulator
	LD	A,0FFH		; set flag
	LD	(LPFLG),A	;
	JR	INTRET		; & return
;
; mnemonics tables
;
; registers
;
MNETAB:	.BYTE	"A "		;
	.WORD	ALOC		;
	.BYTE	"F "		;
	.WORD	FLOC		;
	.BYTE	"H "		;
	.WORD	HLOC		;
	.BYTE	"L "		;
	.WORD	LLOC		;
	.BYTE	"B "		;
	.WORD	BLOC		;
	.BYTE	"C "		;
	.WORD	CLOC		;
	.BYTE	"D "		;
	.WORD	DLOC		;
	.BYTE	"E "		;
	.WORD	ELOC		;
	.BYTE	"I",'X'+80H	;
	.WORD	IXLOC		;
	.BYTE	"I",'Y'+80H	;
	.WORD	IYLOC		;
	.BYTE	"I "		;
	.WORD	ILOC		;
	.BYTE	"IF"		;
	.WORD	IFLOC		;
	.BYTE	"S",'P'+80H	;
	.WORD	SPLOC		;
	.BYTE	"P",'C'+80H	;
	.WORD	PCLOC		;
	.BYTE	"A",27H		;
	.WORD	AALOC		;
	.BYTE	"F",27H		;
	.WORD	FALOC		;
	.BYTE	"H",27H		;
	.WORD	HALOC		;
	.BYTE	"L",27H		;
	.WORD	LALOC		;
	.BYTE	"B",27H		;
	.WORD	BALOC		;
	.BYTE	"C",27H		;
	.WORD	CALOC		;
	.BYTE	"D",27H		;
	.WORD	DALOC		;
	.BYTE	"E",27H		;
	.WORD	EALOC		;
;
; I/O channels
;
	.BYTE	"C",'I'+80H	;
	.WORD	CIADR		;
	.BYTE	"C",'O'+80H	;
	.WORD	COADR		;
	.BYTE	"O",'I'+80H	;
	.WORD	OIADR		;
	.BYTE	"O",'O'+80H	;
	.WORD	OOADR		;
	.BYTE	"S",'I'+80H	;
	.WORD	SIADR		;
	.BYTE	"S",'O'+80H	;
	.WORD	SOADR		;
;
; I/O driver
;
	.BYTE	"TK"		; terminal keyboard
	.WORD	TK		;
	.BYTE	"TT"		; terminal type
	.WORD	TT		;
	.BYTE	"ST"		; Silent 700 printer
	.WORD	ST		;
	.BYTE	"PR"		; papertape reader
	.WORD	PR		;
	.BYTE	"PP"		; papertape punch
	.WORD	PP		;
	.BYTE	"TR"		; 
	.WORD	TR		;
	.BYTE	"AS"		; assembler
	.WORD	AS		;
	.BYTE	"ED"		; editor entry
	.WORD	ED		;
	.BYTE	"LP"		; editor
	.WORD	LP		;
	.BYTE	"ER"		; editor re-entry
	.WORD	ER		;
	.BYTE	"TI"		;
	.WORD	TI		;
	.BYTE	"TO"		;
	.WORD	TO		;
	.BYTE	80H		; end of table
;
	.BYTE	00H,00H,0BFH	;
;
	.END
;
